@@ -24,7 +24,7 @@
  # DAMAGES.
 
 
-__version__ = '0.1.6'
+__version__ = '0.1.7'
 __author__ = 'Jeremy R. Gray'
 __contact__ = 'jrgray@gmail.com'
 
@@ -47,50 +47,60 @@ import hashlib
 from functools import partial  # for buffered hash digest
 from base64 import b64encode, b64decode
 import getpass  # for RSA key-gen
-
+import argparse
 
 lib_name = 'pyFileSec'
 lib_path = abspath(__file__)
-usage = """%(name)s v%(version)s
 
-  File privacy and integrity for psychology and human neuroscience research:
-    encrypt, decrypt, sign, verify, rotate, generate passwords (RSA keys),
-    secure delete, pad (to obscure file-size), & archive.
-    Requires Python 2.6+, and OpenSSL 0.9.8 or higher.
 
-  Module example:
-    >>> import %(name)s as pfs
-    >>> pfs.encrypt('data.txt', 'pub.pem')
-    /path/to/data.enc
-    >>> pfs.decrypt('/path/to/data.enc', 'priv.pem')
-    /path/to/data.txt
+def _parse_args():
+    """Parse and return command line arguments.
 
-  Command-line example:
-    $ alias pfs='%(lib_path)s'
-    $ pfs encrypt data.txt pub.pem
-    /path/to/data.enc
-    $ pfs decrypt data.enc priv.pem
-    /path/to/data.txt
+    a file name is typically the first (required) argument
+    passphrases for command-line usage must go through files;
+        will get a logging.warning()
+    currently not possible to register a new enc/dec method via command line
+    """
+    parser = argparse.ArgumentParser(
+        description='File-oriented privacy & integrity management library.',
+        epilog="See https://pypi.python.org/pypi/pyFileSec/")
+    parser.add_argument('filename', help='path to file to process, "genrsa", or "debug"')
+    parser.add_argument('--version', action='version', version=__version__)
+    parser.add_argument('--verbose', action='store_true', help='print logging info to stdout')
 
-  Options:
-    --help | -h : display this message
-    --version   : print version and exit
-    --openssl=/path/to/openssl : binary file to use for openssl calls
-    --genrsa    : RSA key generation dialog
+    group = parser.add_mutually_exclusive_group()
+    group.add_argument('--encrypt', action='store_true', help='encrypt with RSA + AES256 (-u [-o][-m][-n][-c][-z][-e][-k])')
+    group.add_argument('--decrypt', action='store_true', help='use private key to decrypt (-v [-o][-d][-r])')
+    group.add_argument('--rotate', action='store_true', help='rotate the encryption (-v -U [-V][-r][-R][-z][-e][-c])')
+    group.add_argument('--sign', action='store_true', help='sign file / make signature (-v [-r])')
+    group.add_argument('--verify', action='store_true', help='verify a signature using public key (-u -g)')
+    group.add_argument('--pad', action='store_true', help='obscure file length by padding with bytes ([-z])')
+    group.add_argument('--destroy', action='store_true', help='secure delete')
 
-  Testing:
-    $ py.test %(name)s.py
-  """ % {'name': os.path.splitext(os.path.basename(__file__))[0],
-         'lib_path': lib_path,
-         'version': __version__}
+    parser.add_argument('--openssl', help='specify path of the openssl binary to use')
+    parser.add_argument('-o', '--out', help='path name for generated (output) file')
+    parser.add_argument('-u', '--pub', help='path to public key (.pem file)')
+    parser.add_argument('-U', '--pubn', help='path to new public key (.pem file)')
+    parser.add_argument('-v', '--priv', help='path to private key (.pem file)')
+    parser.add_argument('-V', '--nprv', help='path to new private key (--rotate only)')
+    parser.add_argument('-r', '--pphr', help='path to file containing passphrase for private key')
+    parser.add_argument('-R', '--nppr', help='path to file containing passphrase for new priv key')
+    parser.add_argument('-m', '--meta', help='False to suppress saving meta-data with encrypted file')
+    parser.add_argument('-c', '--hmac', help='path to file containing hmac key')
+    parser.add_argument('-s', '--sig', help='path to signature file (input file for --verify)')
+    parser.add_argument('-z', '--size', type=int, help='num bytes for --pad, default 16384; unpad = 0 or -1')
+    parser.add_argument('-n', '--nodate', action='store_true', help='do not save encryption date in the meta-data')
+    parser.add_argument('-k', '--keep', action='store_true', help='do not --destroy plain-text file after encryption')
+    #parser.add_argument('-e', '--enc', help='register encryption m, nargs=1ethod to use')
+    #parser.add_argument('-d', '--dec', help='register decryption method to use')
+
+    return parser.parse_args()
+
+if __name__ == "__main__":
+    args = _parse_args()
+else:
+    args = None
 
-if '--version' in sys.argv:
-    print(__version__)
-    sys.exit()
-if (__name__ == "__main__" and len(sys.argv) == 1 or
-    '-h' in sys.argv or '--help' in sys.argv):
-    print(usage)
-    sys.exit()
 if python_version() < '2.6.6':
     raise RuntimeError('Requires python 2.6+')
 if python_version < '3.':
@@ -258,9 +268,7 @@ def _setup_logging():
         error = warning = exp = data = info = debug
 
     logging_t0 = get_time()
-    verbose = bool('--verbose' in sys.argv or '--debug' in sys.argv)
-    if '--verbose' in sys.argv:
-        del(sys.argv[sys.argv.index('--verbose')])
+    verbose = args and bool(args.verbose or args.filename == 'debug')
     if not verbose:
         logging = _no_logging()
     else:
@@ -295,7 +303,15 @@ def _sys_call(cmdList, stderr=False, stdin=''):
 def _get_openssl_info():
     """Find, check, and report info about the OpenSSL binary on this system.
     """
-    if sys.platform in ['win32']:
+    if args and args.openssl:
+        OPENSSL = args.openssl
+        logging.info('Option requesting openssl executable: ' + OPENSSL)
+    elif sys.platform not in ['win32']:
+        OPENSSL = _sys_call(['which', 'openssl'])
+        if OPENSSL not in ['/usr/bin/openssl']:
+            msg = 'unexpected location for openssl binary: %s' % OPENSSL
+            logging.warning(msg)
+    else:
         # use a bat file for openssl.cfg; create .bat if not found or broken
         # might have to move to os.environ['HOME'] to ensure write permission
         libdir = os.path.split(os.path.abspath(__file__))[0]
@@ -326,23 +342,9 @@ def _get_openssl_info():
                 with open(OPENSSL, 'wb') as fd:
                     fd.write(where_bat)
         logging.info('will use .bat file for OpenSSL: %s' % OPENSSL)
-    for i, arg in enumerate(sys.argv):
-        if arg.startswith('--openssl='):
-            # spaces in path ok if parses as one argument
-            OPENSSL = arg.replace('--openssl=', '')
-            del(sys.argv[i])
-            logging.info('Option requesting OpenSSL version: ' + OPENSSL)
-            break
-    else:
-        if sys.platform not in ['win32']:
-            OPENSSL = _sys_call(['which', 'openssl'])
-            if OPENSSL not in ['/usr/bin/openssl']:
-                msg = 'unexpected location for openssl binary: %s' % OPENSSL
-                logging.warning(msg)
-        #else:
-        #    OPENSSL = bat_file  # already set
+
     if not isfile(OPENSSL):
-        msg = 'Could not find openssl, tried: %s' % OPENSSL
+        msg = 'Could not find openssl executable, tried: %s' % OPENSSL
         _fatal(msg, RuntimeError)
 
     openssl_version = _sys_call([OPENSSL, 'version'])
@@ -452,6 +454,8 @@ def _entropy_check():
             e = securityd + ' running'
         else:
             e = ''
+        rdrand = _sys_call(['sysctl', 'hw.optional.rdrand'])
+        e += '; rdrand: ' + rdrand
     elif sys.platform.startswith('linux'):
         avail = _sys_call(['cat', '/proc/sys/kernel/random/entropy_avail'])
         e = 'entropy_avail: ' + avail
@@ -743,15 +747,14 @@ def pad(filename, size=16384, test=False, strict=True):
     file sizes exceeding `size` (and hence leak the size of the original file),
     use `strict=False`.
 
-    Padding format: file + n bytes + pad=10-digits + byte + PFS_PAD + byte
+    Padding format: file + n bytes + pad = 10-digits + byte + PFS_PAD + byte
     n is selected to make the new file size == `size`.
 
     `test` allows for testing whether the `size` is adequate to obscure
     the file size. This is similar to testing getsize(file) > size,
     except that `test` also takes into account the padding-size info that is
-    stored as part of the padding (36 bytes, unless max file size is not the
-    default of 1G). So its getsize(file) > size - 36. Testing succeeded
-    if no PaddingError is raised.
+    stored as part of the padding (36 bytes). So its
+    getsize(file) > size - 36. Testing succeeded if no PaddingError is raised.
 
     To make unpadding easier and more robust (= facilitate human inspection),
     the end bytes provide the number of padding bytes that were added, plus an
@@ -1336,7 +1339,7 @@ def rotate(data_enc, priv_old, pub_new, pphr_old=None,
     return new_enc
 
 
-def sign(filename, priv, pphr=None):
+def sign(filename, priv, pphr=None, out=None):
     """Sign a given file with a private key, via `openssl dgst`.
 
     Get a digest of the file, sign the digest, return base64-encoded signature.
@@ -1360,11 +1363,17 @@ def sign(filename, priv, pphr=None):
         _sys_call(cmd_SIGN)
     sig = open(sig_out, 'rb').read()
 
+    if out:
+        with open(out, 'wb)') as fd:
+            fd.write(b64encode(sig))
+        return out
     return b64encode(sig)
 
 
 def verify(filename, pub, sig):
-    """Verify signature of filename using pubkey; expects a base64-encoded sig.
+    """Verify signature of filename using pubkey
+
+    `sig` should be a base64-encoded signature, or a path to a signature file.
     """
     name = 'verify'
     logging.debug(name + ': start, file ' + filename)
@@ -1373,6 +1382,8 @@ def verify(filename, pub, sig):
     else:
         raise NotImplementedError
 
+    if isfile(sig):
+        sig = open(sig, 'rb').read()
     with NamedTemporaryFile(delete=False) as sig_file:
         sig_file.write(b64decode(sig))
     result = _sys_call(cmd_VERIFY + ['-signature', sig_file.name, filename])
@@ -1433,7 +1444,7 @@ def _genRsa2(pub='pub.pem', priv='priv.pem', pphr=None, bits=2048):
 
 
 def genRsaKeys():
-    """Dialog to generate an RSA key pair, PEM format.
+    """Command line dialog to generate an RSA key pair, PEM format.
 
     Limited to 2048, 4096, 8192 bits; 1024 is not secure for medium-term
     storage, and 16384 bits is impractical. Require or generate a passphrase.
@@ -1452,18 +1463,24 @@ def genRsaKeys():
         except:
             pass
 
-    pub = abspath(_uniq_file('pub_RSA.pem'))  # ensure unique, abspath
-    priv = pub.replace('pub_RSA', 'priv_RSA')  # ensure a matched pair
+    # use args for filenames if given explicitly:
+    pub = args.pub or abspath(_uniq_file('pub_RSA.pem'))  # ensure unique
+    priv = args.priv or pub.replace('pub_RSA', 'priv_RSA')  # matched pair
+    pub = abspath(pub)
+    priv = abspath(priv)
+    if pub == priv:
+        priv += '_priv.pem'
+
     if os.path.exists(priv):
         msg = ('RSA key generation.\n  %s already exists\n' % priv +
                '  > Clean up files and try again. Exiting. <')
         print(msg)
         return None, None
 
-    print('RSA key generation. Will try to create two files:')
-    pub_msg = '  %s   = %s' % (os.path.split(pub)[1], pub)
+    print('\nRSA key generation. Will try to create two files:')
+    pub_msg = '  pub  = %s' % pub #os.path.split(pub)#[1]
     print(pub_msg)
-    priv_msg = '  %s  = %s' % (os.path.split(priv)[1], priv)
+    priv_msg = '  priv = %s' % priv #os.path.split(priv)#[1]
     print(priv_msg)
     print('To proceed, enter a passphrase (16+ chars, return to generate).')
     pphr = getpass.getpass('Passphrase: ')
@@ -1517,6 +1534,10 @@ def genRsaKeys():
     return pub, priv
 
 
+def getVersion():
+    return tuple(map(int, __version__.split('.')))
+
+
 class Tests(object):
     """Test suite for py.test
 
@@ -1626,7 +1647,7 @@ class Tests(object):
 
             # test whether encrypt can handle it:
             pub, priv, pphr = self._known_values()[:3]
-            enc = encrypt(filename, pub)  # tarfile fails here with bad filename
+            enc = encrypt(filename, pub)  # tarfile fails here, bad filename
             assert isfile(enc)
 
             dec = decrypt(enc, priv, pphr)
@@ -1974,14 +1995,14 @@ class Tests(object):
         datafile = abspath(datafile)
 
         # Encrypt:
-        cmdLineCmd = [sys.executable, pathToSelf, 'encrypt', datafile,
-                      pub1, '--openssl=' + OPENSSL]
+        cmdLineCmd = [sys.executable, pathToSelf, datafile, '--encrypt',
+                      '--pub', pub1, '--openssl=' + OPENSSL]
         dataEnc = _sys_call(cmdLineCmd).strip()
         assert os.path.isfile(dataEnc)
 
         # Decrypt:
-        cmdLineCmd = [sys.executable, pathToSelf, 'decrypt', dataEnc,
-                      priv1, '--openssl=' + OPENSSL]
+        cmdLineCmd = [sys.executable, pathToSelf, dataEnc, '--decrypt',
+                      '--priv', priv1, '--openssl=' + OPENSSL]
         dataEncDec_cmdline = _sys_call(cmdLineCmd).strip()
         assert os.path.isfile(dataEncDec_cmdline)
 
@@ -2239,6 +2260,7 @@ class Tests(object):
 logging, logging_t0 = _setup_logging()
 if not user_can_link:
     logging.warning('%s: No admin priv; cannot check hardlinks' % lib_name)
+
 OPENSSL, openssl_version, use_rsautl = _get_openssl_info()
 destroy_TOOL, destroy_OPTS = _get_destroy_info()
 
@@ -2248,12 +2270,16 @@ codec = PFSCodecRegistry(default_codec)
 
 if __name__ == '__main__':
     logging.info("%s with %s" % (lib_name, openssl_version))
-    if '--debug' in sys.argv:
-        """Run tests with verbose logging.
+    if args.filename == 'debug':
+        """Run tests with verbose logging; check for memory leaks using gc.
             $ python pyfilesec.py --debug > results.txt
         """
         global pytest
         import pytest
+        import gc
+
+        gc.enable()
+        gc.set_debug(gc.DEBUG_LEAK)
 
         os.mkdir('debug_' + lib_name)
         os.chdir('debug_' + lib_name)  # intermediate files get left inside
@@ -2266,22 +2292,68 @@ if __name__ == '__main__':
                 result = test + ' FAILED'
                 print(result)
         logging.info("%.4fs for tests" % (get_time() - logging_t0))
-    elif '--genrsa' in sys.argv:
-        """Walk through key generation.
+    elif args.filename == 'genrsa':
+        """Walk through key generation on command line.
         """
         genRsaKeys()
     else:
-        """pass sys.args to encrypt or decrypt
+        """Call requested function with arguments, return result (to stdout)
+
+        Functions: encrypt, decrypt, rotate, pad, sign, verify, destroy
         """
-        logging.info(OPENSSL)
-        if sys.argv[1] in ['enc', 'dec']:
-            sys.argv[1] += 'rypt'
-        if sys.argv[1] in ['encrypt', 'decrypt', 'pad', 'unpad']:
-            cmd = sys.argv[1]
-            if cmd == 'unpad':
-                cmd = '_unpad_strict'
-            del sys.argv[1]
-            result = eval(cmd + '(*sys.argv[1:])')
-            print(result)  # full path to file
+        fxn = None  # becomes the actual function
+        kw = {}  # kwargs for fxn
+
+        # "kw.update()" ==> required args, use kw even though its position-able
+        # "arg and kw.update(arg)" ==> optional args; watch out for value == 0
+
+        # mutually exclusive args.fxn:
+        if args.encrypt:
+            fxn = encrypt
+            # convenience arg: call pad the file prior to encryption
+            if args.size >= -1:
+                pad(filename, size=args.size)
+            kw.update({'pub': args.pub})
+            args.keep and kw.update({'keep': args.keep})
+            args.meta and kw.update({'meta': args.meta})
+            args.hmac and kw.update({'hmac_key': args.hmac})
+        elif args.decrypt:
+            fxn = decrypt
+            kw.update({'priv': args.priv})
+            args.pphr and kw.update({'pphr': args.pphr})
+            args.out and kw.update({'outFile': args.out})
+        elif args.rotate:
+            fxn = rotate
+            kw.update({'priv_old': args.priv})
+            kw.update({'pub_new': args.pubn})
+            args.pphr and kw.update({'pphr_old': args.pphr})
+            args.nppr and kw.update({'pphr_new': args.nppr})
+            args.nprv and kw.update({'priv_new': args.nprv})
+            args.nppr and kw.update({'pphr_new': args.nppr})
+            args.keep and kw.update({'keep': args.keep})
+            args.meta and kw.update({'meta': args.meta})
+            args.hmac and kw.update({'hmac_new': args.hmac})
+            if args.size >= -1:
+                kw.update({'pad_new': args.size})
+        elif args.pad:
+            fxn = pad
+            if args.size >= -1:
+                kw.update({'size': args.size})
+            elif args.size is not None:
+                raise ValueError('bad argument for -z/--size to pad')
+        elif args.sign:
+            fxn = sign
+            kw.update({'priv': args.priv})
+            args.pphr and kw.update({'pphr': args.pphr})
+            args.out and kw.update({'out': args.out})
+        elif args.verify:
+            fxn = verify
+            kw.update({'pub': args.pub})
+            kw.update({'sig': args.sig})
+        elif args.destroy:
+            fxn = destroy
         else:
-            print(usage)
+            raise ValueError('No action requested (command line).')
+
+        result = fxn(args.filename, **kw)
+        print result
